/*
 * Copyright 2010-2015 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */

package org.springframework.amqp.rabbit.core;

import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.instanceOf;
import static org.hamcrest.Matchers.startsWith;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertThat;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static org.mockito.Matchers.any;
import static org.mockito.Matchers.anyBoolean;
import static org.mockito.Matchers.anyString;
import static org.mockito.Mockito.atLeast;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;

import com.rabbitmq.client.AMQP.BasicProperties;
import com.rabbitmq.client.Channel;
import com.rabbitmq.client.Consumer;
import com.rabbitmq.client.Envelope;
import com.rabbitmq.client.GetResponse;
import com.rabbitmq.client.ShutdownSignalException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.junit.After;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.ArgumentCaptor;
import org.mockito.Matchers;
import org.mockito.Mockito;
import org.mockito.invocation.InvocationOnMock;
import org.mockito.stubbing.Answer;

import org.springframework.amqp.AmqpException;
import org.springframework.amqp.AmqpIOException;
import org.springframework.amqp.core.Address;
import org.springframework.amqp.core.AmqpMessageReturnedException;
import org.springframework.amqp.core.Message;
import org.springframework.amqp.core.MessagePostProcessor;
import org.springframework.amqp.core.MessageProperties;
import org.springframework.amqp.core.Queue;
import org.springframework.amqp.core.ReceiveAndReplyCallback;
import org.springframework.amqp.core.ReceiveAndReplyMessageCallback;
import org.springframework.amqp.core.ReplyToAddressCallback;
import org.springframework.amqp.rabbit.connection.CachingConnectionFactory;
import org.springframework.amqp.rabbit.connection.Connection;
import org.springframework.amqp.rabbit.connection.ConnectionFactory;
import org.springframework.amqp.rabbit.connection.ConnectionFactoryUtils;
import org.springframework.amqp.rabbit.connection.RabbitResourceHolder;
import org.springframework.amqp.rabbit.connection.SingleConnectionFactory;
import org.springframework.amqp.rabbit.listener.SimpleMessageListenerContainer;
import org.springframework.amqp.rabbit.listener.adapter.MessageListenerAdapter;
import org.springframework.amqp.rabbit.support.ConsumerCancelledException;
import org.springframework.amqp.rabbit.support.DefaultMessagePropertiesConverter;
import org.springframework.amqp.rabbit.support.MessagePropertiesConverter;
import org.springframework.amqp.rabbit.support.PublisherCallbackChannelImpl;
import org.springframework.amqp.rabbit.test.BrokerRunning;
import org.springframework.amqp.rabbit.test.BrokerTestUtils;
import org.springframework.amqp.support.converter.SimpleMessageConverter;
import org.springframework.amqp.support.postprocessor.GUnzipPostProcessor;
import org.springframework.amqp.support.postprocessor.GZipPostProcessor;
import org.springframework.amqp.utils.SerializationUtils;
import org.springframework.amqp.utils.test.TestUtils;
import org.springframework.beans.DirectFieldAccessor;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.expression.common.LiteralExpression;
import org.springframework.scheduling.concurrent.ThreadPoolTaskExecutor;
import org.springframework.test.annotation.DirtiesContext;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.TransactionException;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.interceptor.TransactionAttribute;
import org.springframework.transaction.interceptor.TransactionInterceptor;
import org.springframework.transaction.support.AbstractPlatformTransactionManager;
import org.springframework.transaction.support.DefaultTransactionStatus;
import org.springframework.transaction.support.SimpleTransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionCallbackWithoutResult;
import org.springframework.transaction.support.TransactionSynchronization;
import org.springframework.transaction.support.TransactionSynchronizationAdapter;
import org.springframework.transaction.support.TransactionSynchronizationManager;
import org.springframework.transaction.support.TransactionSynchronizationUtils;
import org.springframework.transaction.support.TransactionTemplate;
import org.springframework.util.ReflectionUtils;
import org.springframework.util.ReflectionUtils.FieldCallback;
import org.springframework.util.ReflectionUtils.FieldFilter;

/**
 * @author Dave Syer
 * @author Mark Fisher
 * @author Tomas Lukosius
 * @author Gary Russell
 * @author Gunnar Hillert
 * @author Artem Bilan
 */
@ContextConfiguration
@RunWith(SpringJUnit4ClassRunner.class)
@DirtiesContext
public class RabbitTemplateIntegrationTests {

	private static final Log logger = LogFactory.getLog(RabbitTemplateIntegrationTests.class);

	private static final String ROUTE = "test.queue";

	private static final Queue REPLY_QUEUE = new Queue("test.reply.queue");

	private RabbitTemplate template;

	@Autowired
	private RabbitTemplate routingTemplate;

	@Autowired
	private ConnectionFactory cf1;

	@Autowired
	private ConnectionFactory cf2;

	@Before
	public void create() {
		final CachingConnectionFactory connectionFactory = new CachingConnectionFactory();
		connectionFactory.setHost("localhost");
		connectionFactory.setPort(BrokerTestUtils.getPort());
		connectionFactory.setPublisherReturns(true);
		template = new RabbitTemplate(connectionFactory);
		template.setSendConnectionFactorySelectorExpression(new LiteralExpression("foo"));
	}

	@After
	public void cleanup() throws Exception {
		((DisposableBean) template.getConnectionFactory()).destroy();
		this.brokerIsRunning.removeTestQueues();
	}

	@Rule
	public BrokerRunning brokerIsRunning = BrokerRunning.isRunningWithEmptyQueues(ROUTE, REPLY_QUEUE.getName());

	@Test
	public void testReceiveNonBlocking() throws Exception {
		this.template.convertAndSend(ROUTE, "nonblock");
		int n = 0;
		String out = (String) this.template.receiveAndConvert(ROUTE);
		while (n++ < 100 && out == null) {
			Thread.sleep(100);
		}
		assertNotNull(out);
		assertEquals("nonblock", out);
		assertNull(this.template.receive(ROUTE));
	}

	@Test(expected = ConsumerCancelledException.class)
	public void testReceiveConsumerCanceled() throws Exception {
		ConnectionFactory connectionFactory = new SingleConnectionFactory("localhost", BrokerTestUtils.getPort());

		class MockConsumer implements Consumer {

			private final Consumer delegate;

			MockConsumer(Consumer delegate) {
				this.delegate = delegate;
			}

			@Override
			public void handleConsumeOk(String consumerTag) {
				this.delegate.handleConsumeOk(consumerTag);
				try {
					handleCancel(consumerTag);
				}
				catch (IOException e) {
					throw new IllegalStateException(e);
				}
			}

			@Override
			public void handleCancelOk(String consumerTag) {
				this.delegate.handleCancelOk(consumerTag);
			}

			@Override
			public void handleCancel(String consumerTag) throws IOException {
				this.delegate.handleCancel(consumerTag);
			}

			@Override
			public void handleShutdownSignal(String consumerTag, ShutdownSignalException sig) {
				this.delegate.handleShutdownSignal(consumerTag, sig);
			}

			@Override
			public void handleRecoverOk(String consumerTag) {
				this.delegate.handleRecoverOk(consumerTag);
			}

			@Override
			public void handleDelivery(String consumerTag, Envelope envelope, BasicProperties properties, byte[] body)
					throws IOException {
				this.delegate.handleDelivery(consumerTag, envelope, properties, body);
			}

		}

		class MockChannel extends PublisherCallbackChannelImpl {

			public MockChannel(Channel delegate) {
				super(delegate);
			}

			@Override
			public String basicConsume(String queue, Consumer callback)
					throws IOException {
				return super.basicConsume(queue, new MockConsumer(callback));
			}

		}

		Connection connection = spy(connectionFactory.createConnection());
		when(connection.createChannel(anyBoolean())).then(new Answer<Channel>() {

			@Override
			public Channel answer(InvocationOnMock invocation) throws Throwable {
				return new MockChannel((Channel) invocation.callRealMethod());
			}

		});

		DirectFieldAccessor dfa = new DirectFieldAccessor(connectionFactory);
		dfa.setPropertyValue("connection", connection);

		this.template = new RabbitTemplate(connectionFactory);
		this.template.setReceiveTimeout(10000);
		this.template.receive(ROUTE);
	}

	@Test
	public void testReceiveBlocking() throws Exception {
		this.template.convertAndSend(ROUTE, "block");
		this.template.setReceiveTimeout(10000);
		String out = (String) this.template.receiveAndConvert(ROUTE);
		assertNotNull(out);
		assertEquals("block", out);
		this.template.setReceiveTimeout(0);
		assertNull(this.template.receive(ROUTE));
	}

	@Test
	public void testReceiveBlockingNoTimeout() throws Exception {
		this.template.convertAndSend(ROUTE, "blockNoTO");
		this.template.setReceiveTimeout(-1);
		String out = (String) this.template.receiveAndConvert(ROUTE);
		assertNotNull(out);
		assertEquals("blockNoTO", out);
		this.template.setReceiveTimeout(1); // test the no message after timeout path
		assertNull(this.template.receive(ROUTE));
	}

	@Test
	public void testReceiveBlockingTx() throws Exception {
		this.template.convertAndSend(ROUTE, "blockTX");
		this.template.setChannelTransacted(true);
		this.template.setReceiveTimeout(10000);
		String out = (String) this.template.receiveAndConvert(ROUTE);
		assertNotNull(out);
		assertEquals("blockTX", out);
		this.template.setReceiveTimeout(0);
		assertNull(this.template.receive(ROUTE));
	}

	@Test
	public void testReceiveBlockingGlobalTx() throws Exception {
		template.convertAndSend(ROUTE, "blockGTXNoTO");
		RabbitResourceHolder resourceHolder = ConnectionFactoryUtils.getTransactionalResourceHolder(
				this.template.getConnectionFactory(), true);
		TransactionSynchronizationManager.setActualTransactionActive(true);
		ConnectionFactoryUtils.bindResourceToTransaction(resourceHolder, this.template.getConnectionFactory(), true);
		template.setReceiveTimeout(-1);
		template.setChannelTransacted(true);
		String out = (String) template.receiveAndConvert(ROUTE);
		resourceHolder.commitAll();
		resourceHolder.closeAll();
		assertSame(resourceHolder, TransactionSynchronizationManager.unbindResource(template.getConnectionFactory()));
		assertNotNull(out);
		assertEquals("blockGTXNoTO", out);
		this.template.setReceiveTimeout(0);
		assertNull(this.template.receive(ROUTE));
	}

	@Test
	public void testSendToNonExistentAndThenReceive() throws Exception {
		// If transacted then the commit fails on send, so we get a nice synchronous exception
		template.setChannelTransacted(true);
		try {
			template.convertAndSend("", "no.such.route", "message");
			// fail("Expected AmqpException");
		}
		catch (AmqpException e) {
			// e.printStackTrace();
		}
		// Now send the real message, and all should be well...
		template.convertAndSend(ROUTE, "message");
		String result = (String) template.receiveAndConvert(ROUTE);
		assertEquals("message", result);
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testSendAndReceiveWithPostProcessor() throws Exception {
		template.convertAndSend(ROUTE, (Object) "message", new MessagePostProcessor() {

			@Override
			public Message postProcessMessage(Message message) throws AmqpException {
				message.getMessageProperties().setContentType("text/other");
				// message.getMessageProperties().setUserId("foo");
				return message;
			}
		});
		String result = (String) template.receiveAndConvert(ROUTE);
		assertEquals("message", result);
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testSendAndReceive() throws Exception {
		template.convertAndSend(ROUTE, "message");
		String result = (String) template.receiveAndConvert(ROUTE);
		assertEquals("message", result);
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testSendAndReceiveUndeliverable() throws Exception {
		template.setMandatory(true);
		try {
			template.convertSendAndReceive(ROUTE + "xxxxxxxx", "undeliverable");
			fail("expected exception");
		}
		catch (AmqpMessageReturnedException e) {
			assertEquals("undeliverable", new String(e.getReturnedMessage().getBody()));
			assertEquals("NO_ROUTE", e.getReplyText());
		}
		assertEquals(0, TestUtils.getPropertyValue(template, "replyHolder", Map.class).size());
	}

	@Test
	public void testSendAndReceiveTransacted() throws Exception {
		template.setChannelTransacted(true);
		template.convertAndSend(ROUTE, "message");
		String result = (String) template.receiveAndConvert(ROUTE);
		assertEquals("message", result);
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testSendAndReceiveTransactedWithUncachedConnection() throws Exception {
		final SingleConnectionFactory singleConnectionFactory = new SingleConnectionFactory("localhost");
		RabbitTemplate template = new RabbitTemplate(singleConnectionFactory);
		template.setChannelTransacted(true);
		template.convertAndSend(ROUTE, "message");
		String result = (String) template.receiveAndConvert(ROUTE);
		assertEquals("message", result);
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
		singleConnectionFactory.destroy();
	}

	@Test
	public void testSendAndReceiveTransactedWithImplicitRollback() throws Exception {
		template.setChannelTransacted(true);
		template.convertAndSend(ROUTE, "message");
		// Rollback of manual receive is implicit because the channel is
		// closed...
		try {
			template.execute(new ChannelCallback<String>() {

				@Override
				public String doInRabbit(Channel channel) throws Exception {
					// Switch off the auto-ack so the message is rolled back...
					channel.basicGet(ROUTE, false);
					// This is the way to rollback with a cached channel (it is
					// the way the ConnectionFactoryUtils
					// handles it via a synchronization):
					channel.basicRecover(true);
					throw new PlannedException();
				}
			});
			fail("Expected PlannedException");
		}
		catch (Exception e) {
			assertTrue(e.getCause() instanceof PlannedException);
		}
		String result = (String) template.receiveAndConvert(ROUTE);
		assertEquals("message", result);
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testSendAndReceiveInCallback() throws Exception {
		template.convertAndSend(ROUTE, "message");
		final MessagePropertiesConverter messagePropertiesConverter = new DefaultMessagePropertiesConverter();
		String result = template.execute(new ChannelCallback<String>() {

			@Override
			public String doInRabbit(Channel channel) throws Exception {
				// We need noAck=false here for the message to be expicitly
				// acked
				GetResponse response = channel.basicGet(ROUTE, false);
				MessageProperties messageProps = messagePropertiesConverter.toMessageProperties(
						response.getProps(), response.getEnvelope(), "UTF-8");
				// Explicit ack
				channel.basicAck(response.getEnvelope().getDeliveryTag(), false);
				return (String) new SimpleMessageConverter().fromMessage(new Message(response.getBody(), messageProps));
			}
		});
		assertEquals("message", result);
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testReceiveInExternalTransaction() throws Exception {
		template.convertAndSend(ROUTE, "message");
		template.setChannelTransacted(true);
		String result = new TransactionTemplate(new TestTransactionManager())
				.execute(new TransactionCallback<String>() {

					@Override
					public String doInTransaction(TransactionStatus status) {
						return (String) template.receiveAndConvert(ROUTE);
					}
				});
		assertEquals("message", result);
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testReceiveInExternalTransactionAutoAck() throws Exception {
		template.convertAndSend(ROUTE, "message");
		// Should just result in auto-ack (not synched with external tx)
		template.setChannelTransacted(true);
		String result = new TransactionTemplate(new TestTransactionManager())
				.execute(new TransactionCallback<String>() {

					@Override
					public String doInTransaction(TransactionStatus status) {
						return (String) template.receiveAndConvert(ROUTE);
					}
				});
		assertEquals("message", result);
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testReceiveInExternalTransactionWithRollback() throws Exception {
		// Makes receive (and send in principle) transactional
		template.setChannelTransacted(true);
		template.convertAndSend(ROUTE, "message");
		try {
			new TransactionTemplate(new TestTransactionManager()).execute(new TransactionCallback<String>() {

				@Override
				public String doInTransaction(TransactionStatus status) {
					template.receiveAndConvert(ROUTE);
					throw new PlannedException();
				}
			});
			fail("Expected PlannedException");
		}
		catch (PlannedException e) {
			// Expected
		}
		String result = (String) template.receiveAndConvert(ROUTE);
		assertEquals("message", result);
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testReceiveInExternalTransactionWithNoRollback() throws Exception {
		// Makes receive non-transactional
		template.setChannelTransacted(false);
		template.convertAndSend(ROUTE, "message");
		try {
			new TransactionTemplate(new TestTransactionManager()).execute(new TransactionCallback<String>() {

				@Override
				public String doInTransaction(TransactionStatus status) {
					template.receiveAndConvert(ROUTE);
					throw new PlannedException();
				}
			});
			fail("Expected PlannedException");
		}
		catch (PlannedException e) {
			// Expected
		}
		// No rollback
		String result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testSendInExternalTransaction() throws Exception {
		template.setChannelTransacted(true);
		new TransactionTemplate(new TestTransactionManager()).execute(new TransactionCallback<Void>() {

			@Override
			public Void doInTransaction(TransactionStatus status) {
				template.convertAndSend(ROUTE, "message");
				return null;
			}
		});
		String result = (String) template.receiveAndConvert(ROUTE);
		assertEquals("message", result);
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testSendInExternalTransactionWithRollback() throws Exception {
		template.setChannelTransacted(true);
		try {
			new TransactionTemplate(new TestTransactionManager()).execute(new TransactionCallback<Void>() {

				@Override
				public Void doInTransaction(TransactionStatus status) {
					template.convertAndSend(ROUTE, "message");
					throw new PlannedException();
				}
			});
			fail("Expected PlannedException");
		}
		catch (PlannedException e) {
			// Expected
		}
		String result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testAtomicSendAndReceive() throws Exception {
		final CachingConnectionFactory cachingConnectionFactory = new CachingConnectionFactory();
		cachingConnectionFactory.setHost("localhost");
		final RabbitTemplate template = new RabbitTemplate(cachingConnectionFactory);
		template.setRoutingKey(ROUTE);
		template.setQueue(ROUTE);
		ExecutorService executor = Executors.newFixedThreadPool(1);
		// Set up a consumer to respond to our producer
		Future<Message> received = executor.submit(new Callable<Message>() {

			@Override
			public Message call() throws Exception {
				Message message = null;
				for (int i = 0; i < 10; i++) {
					message = template.receive();
					if (message != null) {
						break;
					}
					Thread.sleep(100L);
				}
				assertNotNull("No message received", message);
				template.send(message.getMessageProperties().getReplyTo(), message);
				return message;
			}

		});
		Message message = new Message("test-message".getBytes(), new MessageProperties());
		Message reply = template.sendAndReceive(message);
		assertEquals(new String(message.getBody()), new String(received.get(1000, TimeUnit.MILLISECONDS).getBody()));
		assertNotNull("Reply is expected", reply);
		assertEquals(new String(message.getBody()), new String(reply.getBody()));
		// Message was consumed so nothing left on queue
		reply = template.receive();
		assertEquals(null, reply);
		cachingConnectionFactory.destroy();
	}

	@Test
	public void testAtomicSendAndReceiveExternalExecutor() throws Exception {
		final CachingConnectionFactory connectionFactory = new CachingConnectionFactory();
		connectionFactory.setHost("localhost");
		ThreadPoolTaskExecutor exec = new ThreadPoolTaskExecutor();
		final String execName = "make-sure-exec-passed-in";
		exec.setBeanName(execName);
		exec.afterPropertiesSet();
		connectionFactory.setExecutor(exec);
		final Field[] fields = new Field[1];
		ReflectionUtils.doWithFields(RabbitTemplate.class, new FieldCallback() {

			@Override
			public void doWith(Field field) throws IllegalArgumentException,
					IllegalAccessException {
				field.setAccessible(true);
				fields[0] = field;
			}
		}, new FieldFilter() {

			@Override
			public boolean matches(Field field) {
				return field.getName().equals("logger");
			}
		});
		Log logger = Mockito.mock(Log.class);
		when(logger.isTraceEnabled()).thenReturn(true);

		final AtomicBoolean execConfiguredOk = new AtomicBoolean();

		doAnswer(new Answer<Object>() {

			@Override
			public Object answer(InvocationOnMock invocation) throws Throwable {
				String log = (String) invocation.getArguments()[0];
				if (log.startsWith("Message received") &&
						Thread.currentThread().getName().startsWith(execName)) {
					execConfiguredOk.set(true);
				}
				return null;
			}
		}).when(logger).trace(Mockito.anyString());
		final RabbitTemplate template = new RabbitTemplate(connectionFactory);
		ReflectionUtils.setField(fields[0], template, logger);
		template.setRoutingKey(ROUTE);
		template.setQueue(ROUTE);
		ExecutorService executor = Executors.newFixedThreadPool(1);
		// Set up a consumer to respond to our producer
		Future<Message> received = executor.submit(new Callable<Message>() {

			@Override
			public Message call() throws Exception {
				Message message = null;
				for (int i = 0; i < 10; i++) {
					message = template.receive();
					if (message != null) {
						break;
					}
					Thread.sleep(100L);
				}
				assertNotNull("No message received", message);
				template.send(message.getMessageProperties().getReplyTo(), message);
				return message;
			}

		});
		Message message = new Message("test-message".getBytes(), new MessageProperties());
		Message reply = template.sendAndReceive(message);
		assertEquals(new String(message.getBody()), new String(received.get(1000, TimeUnit.MILLISECONDS).getBody()));
		assertNotNull("Reply is expected", reply);
		assertEquals(new String(message.getBody()), new String(reply.getBody()));
		// Message was consumed so nothing left on queue
		reply = template.receive();
		assertEquals(null, reply);

		assertTrue(execConfiguredOk.get());
		connectionFactory.destroy();
	}

	@Test
	public void testAtomicSendAndReceiveWithRoutingKey() throws Exception {
		final CachingConnectionFactory cachingConnectionFactory = new CachingConnectionFactory();
		cachingConnectionFactory.setHost("localhost");
		final RabbitTemplate template = new RabbitTemplate(cachingConnectionFactory);
		ExecutorService executor = Executors.newFixedThreadPool(1);
		// Set up a consumer to respond to our producer
		Future<Message> received = executor.submit(new Callable<Message>() {

			@Override
			public Message call() throws Exception {
				Message message = null;
				for (int i = 0; i < 10; i++) {
					message = template.receive(ROUTE);
					if (message != null) {
						break;
					}
					Thread.sleep(100L);
				}
				assertNotNull("No message received", message);
				template.send(message.getMessageProperties().getReplyTo(), message);
				return message;
			}

		});
		Message message = new Message("test-message".getBytes(), new MessageProperties());
		Message reply = template.sendAndReceive(ROUTE, message);
		assertEquals(new String(message.getBody()), new String(received.get(1000, TimeUnit.MILLISECONDS).getBody()));
		assertNotNull("Reply is expected", reply);
		assertEquals(new String(message.getBody()), new String(reply.getBody()));
		// Message was consumed so nothing left on queue
		reply = template.receive(ROUTE);
		assertEquals(null, reply);
		cachingConnectionFactory.destroy();
	}

	@Test
	public void testAtomicSendAndReceiveWithExchangeAndRoutingKey() throws Exception {
		final CachingConnectionFactory cachingConnectionFactory = new CachingConnectionFactory();
		cachingConnectionFactory.setHost("localhost");
		final RabbitTemplate template = new RabbitTemplate(cachingConnectionFactory);
		ExecutorService executor = Executors.newFixedThreadPool(1);
		// Set up a consumer to respond to our producer
		Future<Message> received = executor.submit(new Callable<Message>() {

			@Override
			public Message call() throws Exception {
				Message message = null;
				for (int i = 0; i < 10; i++) {
					message = template.receive(ROUTE);
					if (message != null) {
						break;
					}
					Thread.sleep(100L);
				}
				assertNotNull("No message received", message);
				template.send(message.getMessageProperties().getReplyTo(), message);
				return message;
			}

		});
		Message message = new Message("test-message".getBytes(), new MessageProperties());
		Message reply = template.sendAndReceive("", ROUTE, message);
		assertEquals(new String(message.getBody()), new String(received.get(1000, TimeUnit.MILLISECONDS).getBody()));
		assertNotNull("Reply is expected", reply);
		assertEquals(new String(message.getBody()), new String(reply.getBody()));
		// Message was consumed so nothing left on queue
		reply = template.receive(ROUTE);
		assertEquals(null, reply);
		cachingConnectionFactory.destroy();
	}

	@Test
	public void testAtomicSendAndReceiveWithConversion() throws Exception {
		final CachingConnectionFactory cachingConnectionFactory = new CachingConnectionFactory();
		cachingConnectionFactory.setHost("localhost");
		final RabbitTemplate template = new RabbitTemplate(cachingConnectionFactory);
		template.setRoutingKey(ROUTE);
		template.setQueue(ROUTE);
		ExecutorService executor = Executors.newFixedThreadPool(1);
		// Set up a consumer to respond to our producer
		Future<String> received = executor.submit(new Callable<String>() {

			@Override
			public String call() throws Exception {
				Message message = null;
				for (int i = 0; i < 10; i++) {
					message = template.receive();
					if (message != null) {
						break;
					}
					Thread.sleep(100L);
				}
				assertNotNull("No message received", message);
				template.send(message.getMessageProperties().getReplyTo(), message);
				return (String) template.getMessageConverter().fromMessage(message);
			}

		});
		String result = (String) template.convertSendAndReceive("message");
		assertEquals("message", received.get(1000, TimeUnit.MILLISECONDS));
		assertEquals("message", result);
		// Message was consumed so nothing left on queue
		result = (String) template.receiveAndConvert();
		assertEquals(null, result);
		cachingConnectionFactory.destroy();
	}

	@Test
	public void testAtomicSendAndReceiveWithConversionUsingRoutingKey() throws Exception {
		ExecutorService executor = Executors.newFixedThreadPool(1);
		// Set up a consumer to respond to our producer
		Future<String> received = executor.submit(new Callable<String>() {

			@Override
			public String call() throws Exception {
				Message message = null;
				for (int i = 0; i < 10; i++) {
					message = template.receive(ROUTE);
					if (message != null) {
						break;
					}
					Thread.sleep(100L);
				}
				assertNotNull("No message received", message);
				template.send(message.getMessageProperties().getReplyTo(), message);
				return (String) template.getMessageConverter().fromMessage(message);
			}

		});
		String result = (String) template.convertSendAndReceive(ROUTE, "message");
		assertEquals("message", received.get(1000, TimeUnit.MILLISECONDS));
		assertEquals("message", result);
		// Message was consumed so nothing left on queue
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testAtomicSendAndReceiveWithConversionUsingExchangeAndRoutingKey() throws Exception {
		ExecutorService executor = Executors.newFixedThreadPool(1);
		// Set up a consumer to respond to our producer
		Future<String> received = executor.submit(new Callable<String>() {

			@Override
			public String call() throws Exception {
				Message message = null;
				for (int i = 0; i < 10; i++) {
					message = template.receive(ROUTE);
					if (message != null) {
						break;
					}
					Thread.sleep(100L);
				}
				assertNotNull("No message received", message);
				template.send(message.getMessageProperties().getReplyTo(), message);
				return (String) template.getMessageConverter().fromMessage(message);
			}

		});
		String result = (String) template.convertSendAndReceive("", ROUTE, "message");
		assertEquals("message", received.get(1000, TimeUnit.MILLISECONDS));
		assertEquals("message", result);
		// Message was consumed so nothing left on queue
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testAtomicSendAndReceiveWithConversionAndMessagePostProcessor() throws Exception {
		final CachingConnectionFactory cachingConnectionFactory = new CachingConnectionFactory();
		cachingConnectionFactory.setHost("localhost");
		final RabbitTemplate template = new RabbitTemplate(cachingConnectionFactory);
		template.setRoutingKey(ROUTE);
		template.setQueue(ROUTE);
		ExecutorService executor = Executors.newFixedThreadPool(1);
		// Set up a consumer to respond to our producer
		Future<String> received = executor.submit(new Callable<String>() {

			@Override
			public String call() throws Exception {
				Message message = null;
				for (int i = 0; i < 10; i++) {
					message = template.receive();
					if (message != null) {
						break;
					}
					Thread.sleep(100L);
				}
				assertNotNull("No message received", message);
				template.send(message.getMessageProperties().getReplyTo(), message);
				return (String) template.getMessageConverter().fromMessage(message);
			}

		});
		String result = (String) template.convertSendAndReceive((Object) "message", new MessagePostProcessor() {

			@Override
			public Message postProcessMessage(Message message) throws AmqpException {
				try {
					byte[] newBody = new String(message.getBody(), "UTF-8").toUpperCase().getBytes("UTF-8");
					return new Message(newBody, message.getMessageProperties());
				}
				catch (Exception e) {
					throw new AmqpException("unexpected failure in test", e);
				}
			}
		});
		assertEquals("MESSAGE", received.get(1000, TimeUnit.MILLISECONDS));
		assertEquals("MESSAGE", result);
		// Message was consumed so nothing left on queue
		result = (String) template.receiveAndConvert();
		assertEquals(null, result);
		cachingConnectionFactory.destroy();
	}

	@Test
	public void testAtomicSendAndReceiveWithConversionAndMessagePostProcessorUsingRoutingKey() throws Exception {
		ExecutorService executor = Executors.newFixedThreadPool(1);
		// Set up a consumer to respond to our producer
		Future<String> received = executor.submit(new Callable<String>() {

			@Override
			public String call() throws Exception {
				Message message = null;
				for (int i = 0; i < 10; i++) {
					message = template.receive(ROUTE);
					if (message != null) {
						break;
					}
					Thread.sleep(100L);
				}
				assertNotNull("No message received", message);
				template.send(message.getMessageProperties().getReplyTo(), message);
				return (String) template.getMessageConverter().fromMessage(message);
			}

		});
		String result = (String) template.convertSendAndReceive(ROUTE, (Object) "message", new MessagePostProcessor() {

			@Override
			public Message postProcessMessage(Message message) throws AmqpException {
				try {
					byte[] newBody = new String(message.getBody(), "UTF-8").toUpperCase().getBytes("UTF-8");
					return new Message(newBody, message.getMessageProperties());
				}
				catch (Exception e) {
					throw new AmqpException("unexpected failure in test", e);
				}
			}
		});
		assertEquals("MESSAGE", received.get(1000, TimeUnit.MILLISECONDS));
		assertEquals("MESSAGE", result);
		// Message was consumed so nothing left on queue
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testAtomicSendAndReceiveWithConversionAndMessagePostProcessorUsingExchangeAndRoutingKey() throws Exception {
		ExecutorService executor = Executors.newFixedThreadPool(1);
		// Set up a consumer to respond to our producer
		Future<String> received = executor.submit(new Callable<String>() {

			@Override
			public String call() throws Exception {
				Message message = null;
				for (int i = 0; i < 10; i++) {
					message = template.receive(ROUTE);
					if (message != null) {
						break;
					}
					Thread.sleep(100L);
				}
				assertNotNull("No message received", message);
				template.send(message.getMessageProperties().getReplyTo(), message);
				return (String) template.getMessageConverter().fromMessage(message);
			}

		});
		String result = (String) template.convertSendAndReceive("", ROUTE, "message", new MessagePostProcessor() {

			@Override
			public Message postProcessMessage(Message message) throws AmqpException {
				try {
					byte[] newBody = new String(message.getBody(), "UTF-8").toUpperCase().getBytes("UTF-8");
					return new Message(newBody, message.getMessageProperties());
				}
				catch (Exception e) {
					throw new AmqpException("unexpected failure in test", e);
				}
			}
		});
		assertEquals("MESSAGE", received.get(1000, TimeUnit.MILLISECONDS));
		assertEquals("MESSAGE", result);
		// Message was consumed so nothing left on queue
		result = (String) template.receiveAndConvert(ROUTE);
		assertEquals(null, result);
	}

	@Test
	public void testReceiveAndReplyNonStandardCorrelationNotBytes() {
		this.template.setQueue(ROUTE);
		this.template.setRoutingKey(ROUTE);
		MessageProperties messageProperties = new MessageProperties();
		messageProperties.getHeaders().put("baz", "bar");
		Message message = new Message("foo".getBytes(), messageProperties);
		this.template.send(ROUTE, message);

		this.template.setCorrelationKey("baz");
		boolean received = this.template.receiveAndReply(new ReceiveAndReplyMessageCallback() {

			@Override
			public Message handle(Message message) {
				return new Message("fuz".getBytes(), new MessageProperties());
			}
		});
		assertTrue(received);
		message = this.template.receive();
		assertNotNull(message);
		assertEquals("bar", message.getMessageProperties().getHeaders().get("baz"));
	}

	@Test
	public void testReceiveAndReplyBlocking() {
		testReceiveAndReply(10000);
	}

	@Test
	public void testReceiveAndReplyNonBlocking() {
		testReceiveAndReply(0);
	}

	private void testReceiveAndReply(long timeout) {
		this.template.setQueue(ROUTE);
		this.template.setRoutingKey(ROUTE);
		this.template.convertAndSend(ROUTE, "test");
		template.setReceiveTimeout(timeout);

		boolean received = this.template.receiveAndReply(new ReceiveAndReplyMessageCallback() {

			@Override
			public Message handle(Message message) {
				message.getMessageProperties().setHeader("foo", "bar");
				return message;
			}
		});
		assertTrue(received);

		Message receive = this.template.receive();
		assertEquals("bar", receive.getMessageProperties().getHeaders().get("foo"));

		this.template.convertAndSend(ROUTE, 1);

		received = this.template.receiveAndReply(ROUTE, new ReceiveAndReplyCallback<Integer, Integer>() {

			@Override
			public Integer handle(Integer payload) {
				return payload + 1;
			}
		});
		assertTrue(received);

		Object result = this.template.receiveAndConvert(ROUTE);
		assertTrue(result instanceof Integer);
		assertEquals(2, result);

		this.template.convertAndSend(ROUTE, 2);

		received = this.template.receiveAndReply(ROUTE, new ReceiveAndReplyCallback<Integer, Integer>() {

			@Override
			public Integer handle(Integer payload) {
				return payload * 2;
			}
		}, "", ROUTE);
		assertTrue(received);

		result = this.template.receiveAndConvert(ROUTE);
		assertTrue(result instanceof Integer);
		assertEquals(4, result);

		if (timeout > 0) {
			this.template.setReceiveTimeout(1);
		}
		received = this.template.receiveAndReply(new ReceiveAndReplyMessageCallback() {

			@Override
			public Message handle(Message message) {
				return message;
			}
		});
		assertFalse(received);

		this.template.convertAndSend(ROUTE, "test");
		this.template.setReceiveTimeout(timeout);
		received = this.template.receiveAndReply(new ReceiveAndReplyMessageCallback() {

			@Override
			public Message handle(Message message) {
				return null;
			}
		});
		assertTrue(received);

		this.template.setReceiveTimeout(0);
		result = this.template.receive();
		assertNull(result);

		this.template.convertAndSend(ROUTE, "TEST");
		this.template.setReceiveTimeout(timeout);
		received = this.template.receiveAndReply(new ReceiveAndReplyMessageCallback() {

			@Override
			public Message handle(Message message) {
				MessageProperties messageProperties = new MessageProperties();
				messageProperties.setContentType(message.getMessageProperties().getContentType());
				messageProperties.setHeader("testReplyTo", new Address("", ROUTE));
				return new Message(message.getBody(), messageProperties);
			}

		}, new ReplyToAddressCallback<Message>() {

			@Override
			public Address getReplyToAddress(Message request, Message reply) {
				return (Address) reply.getMessageProperties().getHeaders().get("testReplyTo");
			}

		});
		assertTrue(received);
		result = this.template.receiveAndConvert(ROUTE);
		assertEquals("TEST", result);

		this.template.setReceiveTimeout(0);
		assertEquals(null, this.template.receive(ROUTE));

		this.template.setChannelTransacted(true);

		this.template.convertAndSend(ROUTE, "TEST");
		this.template.setReceiveTimeout(timeout);
		result = new TransactionTemplate(new TestTransactionManager())
				.execute(new TransactionCallback<String>() {

					@Override
					public String doInTransaction(TransactionStatus status) {
						final AtomicReference<String> payloadReference = new AtomicReference<String>();
						boolean received = template.receiveAndReply(new ReceiveAndReplyCallback<String, Void>() {

							@Override
							public Void handle(String payload) {
								payloadReference.set(payload);
								return null;
							}
						});
						assertTrue(received);
						return payloadReference.get();
					}
				});
		assertEquals("TEST", result);
		this.template.setReceiveTimeout(0);
		assertEquals(null, this.template.receive(ROUTE));

		this.template.convertAndSend(ROUTE, "TEST");
		this.template.setReceiveTimeout(timeout);
		try {
			new TransactionTemplate(new TestTransactionManager())
					.execute(new TransactionCallbackWithoutResult() {

						@Override
						public void doInTransactionWithoutResult(TransactionStatus status) {
							template.receiveAndReply(new ReceiveAndReplyMessageCallback() {

								@Override
								public Message handle(Message message) {
									return message;
								}
							}, new ReplyToAddressCallback<Message>() {

								@Override
								public Address getReplyToAddress(Message request, Message reply) {
									throw new PlannedException();
								}
							});
						}
					});
			fail("Expected PlannedException");
		}
		catch (Exception e) {
			assertTrue(e.getCause() instanceof PlannedException);
		}

		assertEquals("TEST", this.template.receiveAndConvert(ROUTE));
		this.template.setReceiveTimeout(0);
		assertEquals(null, this.template.receive(ROUTE));

		template.convertAndSend("test");
		this.template.setReceiveTimeout(timeout);
		try {
			this.template.receiveAndReply(new ReceiveAndReplyCallback<Double, Void>() {

				@Override
				public Void handle(Double message) {
					return null;
				}
			});
			fail("IllegalArgumentException expected");
		}
		catch (Exception e) {
			assertTrue(e.getCause() instanceof IllegalArgumentException);
			assertTrue(e.getCause().getCause() instanceof ClassCastException);
		}

	}

	@Test
	public void testSymmetricalReceiveAndReply() throws InterruptedException, UnsupportedEncodingException {
		this.template.setQueue(ROUTE);
		this.template.setRoutingKey(ROUTE);
		this.template.setReplyAddress(REPLY_QUEUE.getName());
		this.template.setReplyTimeout(10000);

		SimpleMessageListenerContainer container = new SimpleMessageListenerContainer();
		container.setConnectionFactory(this.template.getConnectionFactory());
		container.setQueues(REPLY_QUEUE);
		container.setMessageListener(this.template);
		container.start();

		int count = 10;

		final Map<Double, Object> results = new HashMap<Double, Object>();

		ExecutorService executor = Executors.newFixedThreadPool(10);

		this.template.setCorrelationKey("CorrelationKey");

		for (int i = 0; i < count; i++) {
			executor.execute(new Runnable() {

				@Override
				public void run() {
					Double request = Math.random() * 100;
					Object reply = template.convertSendAndReceive(request);
					results.put(request, reply);
				}
			});
		}

		for (int i = 0; i < count; i++) {
			executor.execute(new Runnable() {

				@Override
				public void run() {
					Double request = Math.random() * 100;
					MessageProperties messageProperties = new MessageProperties();
					messageProperties.setContentType(MessageProperties.CONTENT_TYPE_SERIALIZED_OBJECT);
					Message reply = template.sendAndReceive(new Message(SerializationUtils.serialize(request), messageProperties));
					results.put(request, SerializationUtils.deserialize(reply.getBody()));
				}
			});
		}

		final AtomicInteger receiveCount = new AtomicInteger();

		long start = System.currentTimeMillis();
		do {
			template.receiveAndReply(new ReceiveAndReplyCallback<Double, Double>() {

				@Override
				public Double handle(Double payload) {
					receiveCount.incrementAndGet();
					return payload * 3;
				}
			});
			if (System.currentTimeMillis() > start + 10000) {
				fail("Something wrong with RabbitMQ");
			}
		}
		while (receiveCount.get() < count * 2);


		executor.shutdown();
		assertTrue(executor.awaitTermination(10, TimeUnit.SECONDS));
		container.stop();

		assertEquals(count * 2, results.size());

		for (Map.Entry<Double, Object> entry : results.entrySet()) {
			assertEquals(entry.getKey() * 3, entry.getValue());
		}

		String messageId = UUID.randomUUID().toString();

		MessageProperties messageProperties = new MessageProperties();
		messageProperties.setMessageId(messageId);
		messageProperties.setContentType(MessageProperties.CONTENT_TYPE_TEXT_PLAIN);
		messageProperties.setReplyTo(REPLY_QUEUE.getName());

		template.send(new Message("test".getBytes(), messageProperties));

		template.receiveAndReply(new ReceiveAndReplyCallback<String, String>() {

			@Override
			public String handle(String payload) {
				return payload.toUpperCase();
			}
		});

		Message result = this.template.receive(REPLY_QUEUE.getName());
		assertEquals("TEST", new String(result.getBody()));
		assertEquals(messageId, new String(result.getMessageProperties().getCorrelationId()));
	}

	@Test
	public void testSendAndReceiveFastImplicit() {
		sendAndReceiveFastGuts();
	}

	@Test
	public void testSendAndReceiveFastExplicit() {
		this.template.setReplyAddress(Address.AMQ_RABBITMQ_REPLY_TO);
		sendAndReceiveFastGuts();
	}

	private void sendAndReceiveFastGuts() {
		try {
			this.template.execute(new ChannelCallback<Void>() {

				@Override
				public Void doInRabbit(Channel channel) throws Exception {
					channel.queueDeclarePassive(Address.AMQ_RABBITMQ_REPLY_TO);
					return null;
				}
			});
			SimpleMessageListenerContainer container = new SimpleMessageListenerContainer();
			container.setConnectionFactory(this.template.getConnectionFactory());
			container.setQueueNames(ROUTE);
			final AtomicReference<String> replyToWas = new AtomicReference<String>();
			MessageListenerAdapter messageListenerAdapter = new MessageListenerAdapter(new Object() {

				@SuppressWarnings("unused")
				public Message handleMessage(Message message) {
					replyToWas.set(message.getMessageProperties().getReplyTo());
					return new Message(new String(message.getBody()).toUpperCase().getBytes(),
							message.getMessageProperties());
				}
			});
			messageListenerAdapter.setMessageConverter(null);
			container.setMessageListener(messageListenerAdapter);
			container.start();
			this.template.setQueue(ROUTE);
			this.template.setRoutingKey(ROUTE);
			Object result = this.template.convertSendAndReceive("foo");
			container.stop();
			assertEquals("FOO", result);
			assertThat(replyToWas.get(), startsWith(Address.AMQ_RABBITMQ_REPLY_TO));
		}
		catch (Exception e) {
			assertThat(e.getCause().getCause().getMessage(), containsString("404"));
			logger.info("Broker does not support fast replies; test skipped " + e.getMessage());
		}
	}

	@Test
	public void testReplyCompressionWithContainer() {
		SimpleMessageListenerContainer container = new SimpleMessageListenerContainer(this.template.getConnectionFactory());
		container.setQueueNames(ROUTE);
		MessageListenerAdapter messageListener = new MessageListenerAdapter(new Object() {

			@SuppressWarnings("unused")
			public String handleMessage(String message) {
				return message.toUpperCase();
			}
		});
		messageListener.setReplyPostProcessor(new GZipPostProcessor());
		container.setMessageListener(messageListener);
		container.setReceiveTimeout(100);
		container.afterPropertiesSet();
		container.start();
		try {
			RabbitTemplate template = new RabbitTemplate();
			template.setConnectionFactory(this.template.getConnectionFactory());
			MessageProperties props = new MessageProperties();
			props.setContentType("text/plain");
			Message message = new Message("foo".getBytes(), props);
			Message reply = template.sendAndReceive("", ROUTE, message);
			assertNotNull(reply);
			assertEquals("gzip:UTF-8", reply.getMessageProperties().getContentEncoding());
			GUnzipPostProcessor unzipper = new GUnzipPostProcessor();
			reply = unzipper.postProcessMessage(reply);
			assertEquals("FOO", new String(reply.getBody()));
		}
		finally {
			container.stop();
		}
	}

	@Test
	public void testDegugLogOnPassiveDeclaration() {
		CachingConnectionFactory connectionFactory = new CachingConnectionFactory("localhost");
		Log logger = spy(TestUtils.getPropertyValue(connectionFactory, "logger", Log.class));
		when(logger.isDebugEnabled()).thenReturn(true);
		new DirectFieldAccessor(connectionFactory).setPropertyValue("logger", logger);
		RabbitTemplate template = new RabbitTemplate(connectionFactory);
		final String queueName = UUID.randomUUID().toString();
		final String exchangeName = UUID.randomUUID().toString();
		try {
			template.execute(new ChannelCallback<Void>() {

				@Override
				public Void doInRabbit(Channel channel) throws Exception {
					channel.queueDeclarePassive(queueName);
					return null;
				}
			});
			fail("Expected exception");
		}
		catch (Exception e) {
			assertThat(e, instanceOf(AmqpIOException.class));
			assertThat(e.getCause(), instanceOf(IOException.class));
			assertThat(e.getCause().getCause(), instanceOf(ShutdownSignalException.class));
			assertThat(e.getCause().getCause().getMessage(), containsString("404"));
		}
		try {
			template.execute(new ChannelCallback<Void>() {

				@Override
				public Void doInRabbit(Channel channel) throws Exception {
					channel.exchangeDeclarePassive(exchangeName);
					return null;
				}
			});
			fail("Expected exception");
		}
		catch (Exception e) {
			assertThat(e, instanceOf(AmqpIOException.class));
			assertThat(e.getCause(), instanceOf(IOException.class));
			assertThat(e.getCause().getCause(), instanceOf(ShutdownSignalException.class));
			assertThat(e.getCause().getCause().getMessage(), containsString("404"));
		}
		verify(logger, never()).error(org.mockito.Matchers.any());
		ArgumentCaptor<Object> logs = ArgumentCaptor.forClass(Object.class);
		verify(logger, atLeast(2)).debug(logs.capture());
		boolean queue = false;
		boolean exchange = false;
		for (Object log : logs.getAllValues()) {
			String logMessage = (String) log;
			queue |= (logMessage.contains(queueName) && logMessage.contains("404"));
			exchange |= (logMessage.contains(queueName) && logMessage.contains("404"));
		}
		assertTrue(queue);
		assertTrue(exchange);
		connectionFactory.destroy();
	}

	@Test
	public void testRouting() throws Exception {
		Connection connection1 = mock(Connection.class);
		when(this.cf1.createConnection()).thenReturn(connection1);
		Channel channel1 = mock(Channel.class);
		when(connection1.createChannel(false)).thenReturn(channel1);
		this.routingTemplate.convertAndSend("exchange", "routingKey", "xyz", new MessagePostProcessor() {

			@Override
			public Message postProcessMessage(Message message) throws AmqpException {
				message.getMessageProperties().setHeader("cfKey", "foo");
				return message;
			}
		});
		verify(channel1).basicPublish(anyString(), anyString(), Matchers.anyBoolean(),
				any(BasicProperties.class), any(byte[].class));

		Connection connection2 = mock(Connection.class);
		when(this.cf2.createConnection()).thenReturn(connection2);
		Channel channel2 = mock(Channel.class);
		when(connection2.createChannel(false)).thenReturn(channel2);
		this.routingTemplate.convertAndSend("exchange", "routingKey", "xyz", new MessagePostProcessor() {

			@Override
			public Message postProcessMessage(Message message) throws AmqpException {
				message.getMessageProperties().setHeader("cfKey", "bar");
				return message;
			}
		});
		verify(channel1).basicPublish(anyString(), anyString(), Matchers.anyBoolean(),
				any(BasicProperties.class), any(byte[].class));
	}

	@Test
	public void testSendInGlobalTransactionCommit() throws Exception {
		testSendInGlobalTransactionGuts(false);

		String result = (String) template.receiveAndConvert(ROUTE);
		assertEquals("message", result);
		assertNull(template.receive(ROUTE));
	}

	@Test
	public void testSendInGlobalTransactionRollback() throws Exception {
		testSendInGlobalTransactionGuts(true);

		assertNull(template.receive(ROUTE));
	}

	@SuppressWarnings({"serial", "unchecked"})
	private void testSendInGlobalTransactionGuts(final boolean rollback) throws Exception {
		template.setChannelTransacted(true);
		new TransactionTemplate(new TestTransactionManager()).execute(new TransactionCallback<Void>() {

			@Override
			public Void doInTransaction(final TransactionStatus status) {

				// Emulate Global TX with the mock non-new TransactionStatus
				new TransactionInterceptor() {

					{
						prepareTransactionInfo(null, null, null, null);
					}

					@Override
					protected TransactionInfo prepareTransactionInfo(PlatformTransactionManager tm,
					                                                 TransactionAttribute txAttr,
					                                                 String joinpointIdentification,
					                                                 TransactionStatus status) {
						TransactionInfo txInfo = new TransactionInfo(tm, txAttr, joinpointIdentification);
						txInfo.newTransactionStatus(new SimpleTransactionStatus(false));

						final ThreadLocal<TransactionInfo> transactionInfoHolder =
								(ThreadLocal<TransactionInfo>) TestUtils.getPropertyValue(this,
										"transactionInfoHolder");
						assertNotNull(transactionInfoHolder);
						transactionInfoHolder.set(txInfo);
						TransactionSynchronizationManager.registerSynchronization(
								new TransactionSynchronizationAdapter() {

									@Override
									public void afterCompletion(int status) {
										transactionInfoHolder.remove();
									}
								});
						return txInfo;
					}

				};
				template.convertAndSend(ROUTE, "message");

				if (rollback) {
					TransactionSynchronizationManager.registerSynchronization(
							new TransactionSynchronizationAdapter() {

								@Override
								public void afterCommit() {
									TransactionSynchronizationUtils.triggerAfterCompletion(
											TransactionSynchronization.STATUS_ROLLED_BACK);
								}

							});
				}

				return null;
			}

		});
	}

	@SuppressWarnings("serial")
	private class PlannedException extends RuntimeException {

		public PlannedException() {
			super("Planned");
		}
	}

	@SuppressWarnings("serial")
	private class TestTransactionManager extends AbstractPlatformTransactionManager {

		@Override
		protected void doBegin(Object transaction, TransactionDefinition definition) throws TransactionException {
		}

		@Override
		protected void doCommit(DefaultTransactionStatus status) throws TransactionException {
		}

		@Override
		protected Object doGetTransaction() throws TransactionException {
			return new Object();
		}

		@Override
		protected void doRollback(DefaultTransactionStatus status) throws TransactionException {
		}

	}

	public static class RCFConfig {

		@Bean
		public ConnectionFactory cf1() {
			return mock(ConnectionFactory.class);
		}

		@Bean
		public ConnectionFactory cf2() {
			return mock(ConnectionFactory.class);
		}

		@Bean
		public ConnectionFactory defaultCF() {
			return mock(ConnectionFactory.class);
		}

	}

}
